% [f,iXX] = linftrain1(X[,l]) Train linf (auxiliary function 1)
%
% This is a special, fast implementation of linftrain.m where:
% - linftrain1 sets the regularisation l and computes an inverse Gram matrix
%   required to fit weights W;
% - linftrain2 fits weights W given the Gram matrices and the outputs Y.
% This allows for fast computation of W when l and the inputs are constant
% but we want to solve for different outputs Y.
% Use as follows:
%   [f,iXX] = linftrain1(X,l);           % precompute stuff given inputs X
%   [f,fX,E] = linftrain2(f,X,iXX,Y1);   % linf model for outputs Y1
%   [f,fX,E] = linftrain2(f,X,iXX,Y2);   % linf model for outputs Y2
% etc.
%
% In:
%   X: NxL matrix, N L-dim data points rowwise.
%   l: (nonnegative scalar) regularisation parameter. Default: 0.
% Out:
%   f: (struct) the linear function, with fields:
%      type='linf', regularisation parameter l.
%      Fields for weights W (DxL) and biases w (Dx1) will be set by linftrain2.
%   iXX: a certain NxL matrix related to X.

% Copyright (c) 2011 by Miguel A. Carreira-Perpinan

function [f,iXX] = linftrain1(X,l)

% ---------- Argument defaults ----------
if ~exist('l','var') l = []; f.l = 0; else f.l = l; end;
% ---------- End of "argument defaults" ----------

[N,L] = size(X); f.type = 'linf';

% Inverse Gram matrix
X1 = sum(X,1)'; XX = X'*X - X1*(X1'/N);
if ~isempty(l) XX = XX + spdiags(repmat(l,L,1),0,L,L); end
iXX = X / XX;

