% [f,G,iGG] = rbftrain1(X,C,s,l) Train RBF (auxiliary function 1)
%
% This is a special, fast implementation of rbftrain.m where:
% - rbftrain1 sets the centres C, width s and regularisation l, and computes
%   the Gram and inverse Gram matrices required to fit weights W;
% - rbftrain2 fits weights W given the Gram matrices and the outputs Y.
% This allows for fast computation of W when C,s,l and the inputs are constant
% but we want to solve for different outputs Y.
% Use as follows:
%   [f,G,iGG] = rbftrain1(X,C,s,l);     % precompute stuff given inputs X
%   [f,fX,E] = rbftrain2(f,G,iGG,Y1);   % RBF model for outputs Y1
%   [f,fX,E] = rbftrain2(f,G,iGG,Y2);   % RBF model for outputs Y2
% etc.
%
% In:
%   X: NxL matrix, N L-dim data points rowwise.
%   C: RBF centres, one of these 3 options:
%      - MxL matrix: the M L-dim values of the BF centres rowwise to use.
%      - cell {M[,R]} where M and optional R are positive integers: the number
%        M of BFs to use, learned by running k-means R times with random
%        initialisation and picking the best (default: 10);
%      - cell {C[,R]} where C is an MxL matrix containing M L-dim initial
%        values of the BF centres rowwise; k-means is then ran once from this
%        initial C for a maximum number of R iterations (default: Inf).
%   s: (positive scalar) RBF width.
%   l: (nonnegative scalar) regularisation parameter.
% Out:
%   f: (struct) the RBF, with fields:
%      type='rbf', centres C (MxL), width s, regularisation parameter l.
%      Fields for weights W (DxM) and biases w (Dx1) will be set by rbftrain2.
%   G: MxN Gram matrix with Gaussian RBF values for X.
%   iGG: a certain NxM matrix related to G.

% Copyright (c) 2011 by Miguel A. Carreira-Perpinan

function [f,G,iGG] = rbftrain1(X,C,s,l)

% ---------- Argument defaults ----------
if iscell(C) & (length(C)==1)
  if isscalar(C{1}) C{2} = 10; else C{2} = Inf; end;
end
% ---------- End of "argument defaults" ----------

N = size(X,1); M = size(C,1); f.type = 'rbf';

% Fit centres
if iscell(C)
  if isscalar(C{1})
    M = C{1}; e = Inf;	% run kmeans C{2} times with rnd init and pick best
    for i=1:C{2}
      [tmpC,~,tmpe] = kmeans(X,M);
      if tmpe(end) < e C = tmpC; e = tmpe(end); end
    end
  else			% run kmeans once from C{1} for at most C{2} iterations
    C = kmeans(X,size(C{1},1),C{1},C{2});
  end
end
f.C = C;

% Set width and regularisation parameter
f.s = s; f.l = l;

% Gram matrices
%G = exp(-sqdist(C,X)/(2*s*s));		% Gram matrix (centres,inputs)
G = dexp(-sqdist(C,X)/(2*s*s));
G1 = sum(G,2); iGG = G' / (G*G' - G1*(G1'/N) + spdiags(repmat(l,M,1),0,M,M));

