% [ZZ,FF,gg,E]=drsvm(X,Y,Z,Fmap,gmap[,beta,tol,maxit])
%   Dimension reduction for Support Vector Machines. The classifier is a
%   combination of two mappings: a dimension reduction mapping F, and a
%   linear SVM g, i.e., X-->Z=F(X)-->Y=g(Z).
%
% Typical use with RBFs:
%   Fmap={'rbf', {70}, 1.0, 1e-5};
%   gmap={'svm', 1};
%   [ZZ,FF,gg,E]=drsvm(X,Y,Z,Fmap,gmap,beta,[],30);
% If you want to give specific initial values to the RBF centres or fix
% their width, note carefully how 'rbftrain' is called below.
%
% In:
%   X: NxD matrix containing N D-dim data points rowwise (input data).
%   Y: NxM matrix containing N M-dim data points rowwise (output label).
%   Z: NxL matrix containing N L-dim data points rowwise (initial low-dim
%      points, obtained from a different method, e.g. KLDA).
%   Fmap, gmap: cell arrays describing the type of mapping (F:X->Z, g:Z->Y):
%    F={'rbf',C,s,l}: RBF network with parameters C,s,l as in 'rbftrain'.
%    F={'linf',l}: linear mapping with regularization parameter l.
%    g={'svm',C}: linear SVM with penalty parameter C.
%   beta: starting penalty parameter.
%   tol: minimum relative decrease in error to keep iterating. Default: 1e-4.
%   maxit: maximum number of iterations. Default: 1000.
% Out:
%   H: the final classifier: g(F(X)), the same type as F.
%   E: training errors for each iteration (scalar).
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.
%
% Copyright (c) 2013 by Weiran Wang and Miguel Carreira-Perpinan.

function [H,E]=drsvm(X,YY,Z,Fmap,gmap,beta,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('beta','var') || isempty(beta)
  beta=1;
end
if ~exist('tol','var') || isempty(tol)
  tol=1e-5;
end
if ~exist('maxit','var') || isempty(maxit)
  maxit=1000;
end
for i=length(Fmap)+1:4
  Fmap{i}=[];
end
% ---------- End of "argument defaults" ----------

% Normalizing Z.
mZ=mean(Z,1);  Z=bsxfun(@minus,Z,mZ);  sZ=norm(Z,'fro');  Z=Z/sZ;

% Number of One-Versus-All SVM classifier.
[~,M]=size(YY);

% Fit mappings g, F.
i=0;  EG=0;  G=cell(1,M);  svm_options=['-q -s 0 -t 0 -c ' num2str(gmap{2})];
parfor j=1:M  % One-Versus-All scheme.
  Yj=YY(:,j);  g=svmtrain(Yj,Z,svm_options);
  g.c=gmap{2}; g.normalw=g.sv_coef'*g.SVs;
  g.bias=-g.rho;  Y2=Z*g.normalw'+g.bias;  g_slacks=max(1-Yj.*Y2,0);
  Eg=0.5*sum(g.normalw.^2)+g.c*sum(g_slacks);  EG=EG+Eg;  G{j}=g;
end
switch Fmap{1}
  case 'rbf',
    % Cache partial solution for fixed penalty parameter.
    [f0,FG,FiGG]=rbftrain1(X,Fmap{2:3},Fmap{4}/beta);
    [F,FX,EF]=rbftrain2(f0,FG,FiGG,Z);  EF=beta*sum(EF);
  case 'linf',
    % Cache partial solution for fixed penalty parameter.
    [f0,FiGG]=linftrain1(X,Fmap{2}/beta);
    [F,FX,EF]=linftrain2(f0,X,FiGG,Z);  EF=beta*sum(EF);
end
E=EG+EF;  H=assemble(F,G);  bias=norm(Z-FX,'fro');

newbeta=0;  % Keep track of the update of penalty parameter.

while i<maxit
  i=i+1;  fprintf('Iteration  %03d: ',i);
  
  % PROJECTION: optimization over Z.
  Z=Zopt(Z,YY,G,FX,beta);
  EG=0;
  for j=1:M
    g=G{j};  Yj=YY(:,j);  Y2=Z*G{j}.normalw'+G{j}.bias;
    EG=EG+0.5*sum(g.normalw.^2)+g.c*sum(max(1-Yj.*Y2,0));
  end
  E=[E, EG+EF];  fprintf('beta=%g, error=%g\n',beta,E(end));
  
  % ADAPTATION: optimization over g, F.
  EG=0;  G=cell(1,M);  svm_options=['-q -e 0.0001 -s 0 -t 0 -c ' num2str(gmap{2})];
  parfor j=1:M   % One-Versus-All scheme.
    Yj=YY(:,j);  g=svmtrain(Yj,Z,svm_options);
    g.c=gmap{2}; g.normalw=g.sv_coef'*g.SVs;
    g.bias=-g.rho;  Y2=Z*g.normalw'+g.bias;  g_slacks=max(1-Yj.*Y2,0);
    Eg=0.5*sum(g.normalw.^2)+g.c*sum(g_slacks);  EG=EG+Eg;  G{j}=g;
  end
  switch Fmap{1}
    case 'rbf',
      if newbeta==1  % Recompute partial solution if beta has changed.
        [f0,FG,FiGG]=rbftrain1(X,Fmap{2:3},Fmap{4}/beta);
      end
      [F,FX,EF]=rbftrain2(f0,FG,FiGG,Z);  EF=beta*sum(EF);
    case 'linf',
      if newbeta==1  % Recompute partial solution if beta has changed.
        [f0,FiGG]=linftrain1(X,Fmap{2}/beta);
      end
      [F,FX,EF]=linftrain2(f0,X,FiGG,Z);  EF=beta*sum(EF);
  end
  H=assemble(F,G);  bias=norm(Z-FX,'fro');
  
  if newbeta==1
    newbeta=0;  % Update beta after at least 1 iteration.
  else
    % Increase beta when convergence is achieved for fixed beta.
    if (abs(E(end-1)-E(end)) < tol*abs(E(end-1)))
      beta=1.5*beta;  newbeta=1;
      % Finish the program if constraints are satisfied.
      if bias < tol*norm(Z,'fro')
        break;
      end
    end
  end
  
end
fprintf('Done.\n');



% Z-step: solving a quadratic program.
function [Z,slacks]=Zopt(Z,YY,G,FX,beta)

[N,L]=size(Z);  M=size(YY,2);  slacks=zeros(N,M);

if M==1
  % Binary classification problem, close-form solution.
  w=G{1}.normalw;  b=G{1}.bias;  c=G{1}.c/beta;
  %% Old implementation.
  %   tic;
  %   LHS=YY.*(FX*w'+b);
  %   % If classified correctly with big margin, z=F(x), slack=0.
  %   idx=find(LHS>=1);  Z(idx,:)=FX(idx,:);
  %   % Otherwise.
  %   idx=find(LHS<1);  lambda=2*(1-LHS(idx))/(w*w');  lambda(lambda>c)=c;
  %   Z(idx,:)=FX(idx,:)+0.5*(lambda.*YY(idx))*w;
  %   slacks(idx)=1-LHS(idx)-0.5*lambda*w*w';
  %   toc;
  %% New implementation.
  tic;
  LAMBDA1=2*max(0,1-YY.*(FX*w'+b))/sum(w.^2);
  Z=FX+(0.5*min(LAMBDA1,c).*YY)*w;
  slacks=0.5*sum(w.^2)*max(0,LAMBDA1-c);
  toc;
else
  % Multi-class problem, call quadratic program solver.
  options=optimset('Display','off','LargeScale','off',...
    'algorithm','active-set','Maxiter',5000,'TolX',1e-12,'TolFun',1e-12);
  parfor i=1:N
    % Setup the problem.
    z=Z(i,:);  Fx=FX(i,:);
    A=[];  b=[];  f=-2*Fx';
    for j=1:M
      A=[A; -YY(i,j)*G{j}.normalw];
      b=[b; -1+YY(i,j)*G{j}.bias];
      f=[f; G{j}.c/beta];
    end
    A=[A, -eye(M); zeros(M,L), -eye(M)];  b=[b; zeros(M,1)];
    H=[2*eye(L), zeros(L,M); zeros(M,L), zeros(M)];
    [z,~,exitflag]=quadprog(H,f,A,b,[],[],[],[],[z,zeros(1,M)]',options);
    if exitflag~=1
      fprintf('z might not be optimal!\n');
    end
    Z(i,:)=z(1:L);  slacks(i,:)=z(L+1:end);
  end
end



% Assemble F (linf/RBF) and G (one-versus-all SVMs) to give final mapping.
% H has the same type as F.
function H=assemble(F,G)

H.type=F.type;
if strcmp(H.type,'rbf')
  H.C=F.C;  H.s=F.s;
end

WF=F.W;  wF=F.w;
M=length(G);  H.W=zeros(M,size(WF,2));  H.w=zeros(M,1);
for i=1:M
  H.W(i,:)=G{i}.normalw*WF;
  H.w(i)=G{i}.normalw*wF+G{i}.bias;
end
